/* client.c */

#include <sys/utsname.h>
#include "fumied.h"

int (*Fsm[MAX_STATES])();		/* finite state machine */
int CurrState;				/* current state */
int PrevState;				/* previous state */

struct fumiedMessage FumiedMsgSend;	/* FUMIED message to send */
struct fumiedMessage FumiedMsgRecv;	/* FUMIED message received */

int Ssend;				/* socket fd for send */
int Srecv;				/* socket fd for receive */

time_t ReqSentTime;		/* time when FUMIEDREQUEST message is sent */

void
fumiedMsgInit(u_char *ifname)
{
  struct sockaddr_in addr;

  setSockAddrIn(htons(FUMIED_CLIENT_PORT), htonl(INADDR_ANY), &addr);
  openSendSocket(&addr, &Ssend);
  Srecv = Ssend;
}

void
fumiedClient()
{
  int next_state;

  Fsm[INIT]       = init;
  Fsm[REQUESTING] = request;
  Fsm[GETTING]    = getinfo;

  CurrState = PrevState = INIT;

  for(;;)
    {
    next_state = (*Fsm[CurrState])();
    if ( next_state == EXCEPTION )
      {
	error("Exception occured in the fsm (fumiedClient)");
      }
    else if ( next_state == COMPLETED )
      {
	note("Request completed");
	return;
      }
    PrevState = CurrState;
    CurrState  = next_state;
    }
}

int
init()
{
  int waitTime = 0;
  struct sockaddr_in	addr;

  mkFumiedDiscoverMsg(&FumiedMsgSend);
  setSockAddrIn(htons(FUMIED_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);

  do
    {
      waitTime = random() % 10;
    }
  while ( waitTime == 0 );
  sleep(waitTime);
  time(&ReqSentTime);	/* record local time
			   not @ REQUEST but @ DISCOVER (RFC1541) */
  if ( sendto(Ssend, (char *)&FumiedMsgSend, sizeof(FumiedMsgSend), 0,
	      (struct sockaddr *)&addr, sizeof(addr)) < 0 )
    {
      error("sendto (init)");
    }
  return REQUESTING;
}

int
request()
{
  struct sockaddr_in addr;
  u_int waitMsgType;
  char local_hostname[80];
  char local_ipaddr[80];
  struct utsname sysname;
  struct in_addr l_addr;

  /* wait FUMIEDACK/FUMIEDNAK, rexmit RHCPREQUEST if necessary
   */
  setSockAddrIn(htons(FUMIED_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);
  waitMsgType = 0;
  setWaitMsgType(FUMIEDACK, &waitMsgType);
  setWaitMsgType(FUMIEDNAK, &waitMsgType);
  if ( waitChkReXmitMsg(Srecv, &FumiedMsgRecv, Ssend, &FumiedMsgSend,
			&addr, waitMsgType, N_REXMIT_REQUEST) )
    {
    mkFumiedRequestMsg(&FumiedMsgSend);
#ifdef NEED_BCAST_RESPONSE
    FumiedMsgSend.flags = htons(F_BROADCAST);
#endif
    setSockAddrIn(htons(FUMIED_SERVER_PORT),
		  (u_int )(FumiedMsgRecv.siaddr.s_addr), &addr);

    if ( sendto(Ssend, (char *)&FumiedMsgSend, sizeof(FumiedMsgSend), 0,
		(struct sockaddr *)&addr, sizeof(addr)) < 0 )
      {
	error("sendto (selecting)");
      }
    return GETTING;
    }
  else
    {
    l_addr.s_addr = Ifbuf.addr;
    strcpy(local_ipaddr, inet_ntoa(l_addr));
    if (!uname(&sysname))
      {
	strcpy(local_hostname, sysname.nodename);
      }
    note("local hostname:%s", local_hostname);
    note("local ipaddr:%s", local_ipaddr);
    revise_hosts(local_hostname, local_ipaddr, NULL, NULL);
    return COMPLETED;
    }
}

int
getinfo()
{
  struct sockaddr_in addr;
  u_int waitMsgType;
  char peer_hostname[80];
  char local_hostname[80];
  char peer_ipaddr[80];
  char local_ipaddr[80];
  struct utsname sysname;
  struct in_addr l_addr;

  /* wait FUMIEDACK/FUMIEDNAK, rexmit RHCPREQUEST if necessary
   */
  setSockAddrIn(htons(FUMIED_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);
  waitMsgType = 0;
  setWaitMsgType(FUMIEDACK, &waitMsgType);
  setWaitMsgType(FUMIEDNAK, &waitMsgType);
  if ( waitChkReXmitMsg(Srecv, &FumiedMsgRecv, Ssend, &FumiedMsgSend,
			&addr, waitMsgType, N_REXMIT_REQUEST) )
    {
    strncpy(peer_hostname, &FumiedMsgRecv.options[5], sizeof(peer_hostname)); 
    strcpy(peer_ipaddr, inet_ntoa(FumiedMsgRecv.siaddr));
    note("peer hostname:%s", peer_hostname);
    note("peer ipaddr:%s", peer_ipaddr);
    l_addr.s_addr = Ifbuf.addr;
    strcpy(local_ipaddr, inet_ntoa(l_addr));
    if (!uname(&sysname))
      {
	strcpy(local_hostname, sysname.nodename);
      }
    note("local hostname:%s", local_hostname);
    note("local ipaddr:%s", local_ipaddr);
    revise_hosts(local_hostname, local_ipaddr, peer_hostname,
		 peer_ipaddr);
    return COMPLETED;
  }
  warn("no response to FUMIEDREEQUEST message. move to INIT state");
  return EXCEPTION;
}

void
mkFumiedDiscoverMsg(struct fumiedMessage *msg)
{
  u_char *p = msg->options + 4; /* just after the magic cookie */

  bzero((char *)msg, sizeof(*msg));
  msg->htype = HTYPE_ETHER;	/* supports Etherenet only */
  msg->hlen = 6;
#ifdef NEED_BCAST_RESPONSE
  msg->flags = htons(F_BROADCAST);
#endif
  msg->ciaddr.s_addr = Ifbuf.addr;

  /* make FUMIED message option field
   */
  *((u_long *)msg->options) = htonl(MAGIC_COOKIE);
  *p++ = FUMIEDDISCOVER;
}

void
mkFumiedRequestMsg(struct fumiedMessage *msg)
{
  u_char *p = msg->options + 4; /* just after the magic cookie */
  struct utsname sysname;

  bzero((char *)msg, sizeof(*msg));
  msg->htype = HTYPE_ETHER;	/* supports Etherenet only */
  msg->hlen = 6;
  msg->ciaddr.s_addr = Ifbuf.addr;

  /* make FUMIED message option field
   */
  *((u_long *)msg->options) = htonl(MAGIC_COOKIE);
  *p++ = FUMIEDREQUEST;
  if (!uname(&sysname))
    {
      strcpy(p, sysname.nodename);
      p += strlen(sysname.nodename);
    }
  *p++ = '\0';
}

long
getNextTimeout(int flag)
{
  static long prevTimeout;
  long rv;			/* return value */

  if ( flag == INIT_TIMEOUT )
    {
      prevTimeout = 2;		/* 2 seconds */
    }
  else if ( prevTimeout > 64 )
    {
      prevTimeout = 2;
    }

  rv = (prevTimeout - 1) * 1000000 + rand()/2000000;
  prevTimeout *= 2;
  return rv;
}
