/* dispatch.c */

#include "fumied.h"
#include <sys/ioctl.h>

struct interface_info *interfaces, *dummy_interfaces;
struct protocol *protocols;
struct timeout *timeouts;
static struct timeout *free_timeouts;
static int interfaces_invalidated;
void (*fumied_packet_handler)(struct interface_info *interface,
			     u_char *packbuf, int len,
			     unsigned short from_port, struct hardware *hfrom);

static void got_one(struct protocol *l);

/* Use the SIOCGIFCONF ioctl to get a list of all the attached interfaces.
   For each interface that's of type INET and not the loopback interface,
   register that interface with the network I/O software, figure out what
   subnet it's on, and add it to the list of interfaces. */

void
discover_interfaces(int state)
{
  struct interface_info *tmp;
  struct interface_info *last, *next;
  char buf [8192];
  struct ifconf ic;
  struct ifreq ifr;
  int i;
  int sock;
  struct sockaddr_in foo;
  int ir;
  char *s;

  /* Create an unbound datagram socket to do the SIOCGIFADDR ioctl on. */
  if ((sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    error ("Can't create addrlist socket");

  /* Get the interface configuration information... */
  ic.ifc_len = sizeof buf;
  ic.ifc_ifcu.ifcu_buf = (caddr_t)buf;
  i = ioctl(sock, SIOCGIFCONF, &ic);

  if (i < 0)
    error ("ioctl: SIOCGIFCONF: %m");

  /* If we already have a list of interfaces, and we're running as
     a FUMIED server, the interfaces were requested. */
  if (interfaces && (state == DISCOVER_SERVER ||
		     state == DISCOVER_RELAY ||
		     state == DISCOVER_REQUESTED))
    ir = 0;
  else if (state == DISCOVER_UNCONFIGURED)
    ir = INTERFACE_REQUESTED | INTERFACE_AUTOMATIC;
  else
    ir = INTERFACE_REQUESTED;

  /* Cycle through the list of interfaces looking for IP addresses.
     Go through twice; once to count the number of addresses, and a
     second time to copy them into an array of addresses. */
  for (i = 0; i < ic.ifc_len;)
    {
      struct ifreq *ifp = (struct ifreq *)((caddr_t)ic.ifc_req + i);
      i += sizeof *ifp;

      if ((s = strrchr (ifp -> ifr_name, ':')))
	{
	  *s = 0;
	}

      if (!strncmp (ifp -> ifr_name, "dummy", 5))
	continue;

      /* See if this is the sort of interface we want to
	 deal with. */
      strcpy (ifr.ifr_name, ifp -> ifr_name);
      if (ioctl (sock, SIOCGIFFLAGS, &ifr) < 0)
	error ("Can't get interface flags for %s: %m", ifr.ifr_name);

      /* Skip loopback, point-to-point and down interfaces,
	 except don't skip down interfaces if we're trying to
	 get a list of configurable interfaces. */
      if ((ifr.ifr_flags & IFF_LOOPBACK) || (ifr.ifr_flags & IFF_POINTOPOINT)
	  || (!(ifr.ifr_flags & IFF_UP) && state != DISCOVER_UNCONFIGURED))
	continue;
		
      /* See if we've seen an interface that matches this one. */
      for (tmp = interfaces; tmp; tmp = tmp -> next)
	if (!strcmp (tmp -> name, ifp -> ifr_name))
	  break;

      /* If there isn't already an interface by this name,
	 allocate one. */
      if (!tmp)
	{
	  tmp = ((struct interface_info *)
		 dmalloc (sizeof *tmp, "discover_interfaces"));
	  if (!tmp)
	    error ("Insufficient memory to %s %s",
		   "record interface", ifp -> ifr_name);
	  strcpy (tmp -> name, ifp -> ifr_name);
	  tmp -> next = interfaces;
	  tmp -> flags = ir;
	  interfaces = tmp;
	}

      if (ifp -> ifr_addr.sa_family == AF_INET)
	{
#ifdef SIOCGIFHWADDR
	  struct ifreq ifr;
	  struct sockaddr sa;

	  /* Read the hardware address from this interface. */
	  ifr = *ifp;
	  if (ioctl (sock, SIOCGIFHWADDR, &ifr) < 0)
	    error ("Can't get hardware address for %s: %m",
		   ifr.ifr_name);

	  sa = *(struct sockaddr *)&ifr.ifr_hwaddr;

	  switch (sa.sa_family)
	    {
#ifdef ARPHRD_LOOPBACK
	    case ARPHRD_LOOPBACK:
				/* ignore loopback interface */
	      break;
#endif

	    case ARPHRD_ETHER:
	      tmp -> hw_address.hlen = 6;
	      tmp -> hw_address.htype = ARPHRD_ETHER;
	      memcpy (tmp -> hw_address.haddr, sa.sa_data, 6);
	      break;

#ifndef ARPHRD_IEEE802
# define ARPHRD_IEEE802 HTYPE_IEEE802
#endif
	    case ARPHRD_IEEE802:
	      tmp -> hw_address.hlen = 6;
	      tmp -> hw_address.htype = ARPHRD_IEEE802;
	      memcpy (tmp -> hw_address.haddr, sa.sa_data, 6);
	      break;

#ifdef ARPHRD_METRICOM
	    case ARPHRD_METRICOM:
	      tmp -> hw_address.hlen = 6;
	      tmp -> hw_address.htype = ARPHRD_METRICOM;
	      memcpy (tmp -> hw_address.haddr, sa.sa_data, 6);
	      break;
#endif

	    default:
	      error ("%s: unknown hardware address type %d",
		     ifr.ifr_name, sa.sa_family);
	    }
#endif /* defined (SIOCGIFHWADDR) */

	  /* Get a pointer to the address... */
	  memcpy (&foo, &ifp -> ifr_addr, sizeof ifp -> ifr_addr);

	  /* We don't want the loopback interface. */
	  if (foo.sin_addr.s_addr == htonl (INADDR_LOOPBACK))
	    continue;

	  /* If this is the first real IP address we've
	     found, keep a pointer to ifreq structure in
	     which we found it. */
	  if (!tmp -> ifp)
	    {
	      struct ifreq *tif;
	      int len = sizeof *ifp;
	      tif = (struct ifreq *)malloc (len);
	      if (!tif)
		error ("no space to remember ifp.");
	      memcpy (tif, ifp, len);
	      tmp -> ifp = tif;
	      tmp -> primary_address = foo.sin_addr;
	    }
	}
    }

  /* If we're just trying to get a list of interfaces that we might
     be able to configure, we can quit now. */
  if (state == DISCOVER_UNCONFIGURED)
    return;

  /* Weed out the interfaces that did not have IP addresses. */
  last = (struct interface_info *)0;
  for (tmp = interfaces; tmp; tmp = next)
    {
      next = tmp -> next;
      if ((tmp -> flags & INTERFACE_AUTOMATIC) && state == DISCOVER_REQUESTED)
	tmp -> flags &= ~(INTERFACE_AUTOMATIC | INTERFACE_REQUESTED);
      if (!tmp -> ifp || !(tmp -> flags & INTERFACE_REQUESTED))
	{
	  if ((tmp -> flags & INTERFACE_REQUESTED) != ir)
	    error ("%s: not found", tmp -> name);
	  if (!last)
	    interfaces = interfaces -> next;
	  else
	    last -> next = tmp -> next;

	  /* Remember the interface in case we need to know
	     about it later. */
	  tmp -> next = dummy_interfaces;
	  dummy_interfaces = tmp;
	  continue;
	}
      last = tmp;

      memcpy (&foo, &tmp -> ifp -> ifr_addr, sizeof tmp -> ifp -> ifr_addr);
      /* Register the interface... */
      if_register_receive (tmp);
      if_register_send (tmp);
    }

  /* Now register all the remaining interfaces as protocols. */
  for (tmp = interfaces; tmp; tmp = tmp -> next)
    add_protocol (tmp -> name, tmp -> rfdesc, got_one, tmp);

  close (sock);
}

/* Wait for packets to come in using select().   When one does, call
   receive_packet to receive the packet and possibly strip hardware
   addressing information from it, and then call do_packet to try to
   do something with it. */

void
dispatch()
{
  fd_set r, w, x;
  struct protocol *l;
  int max = 0;
  int count;
  struct timeval tv, *tvp;

  FD_ZERO (&w);
  FD_ZERO (&x);

  do
    {
      /* Call any expired timeouts, and then if there's
	 still a timeout registered, time out the select
	 call then. */
    another:
      if (timeouts)
	{
	  struct timeout *t;
	  if (timeouts -> when <= cur_time)
	    {
	      t = timeouts;
	      timeouts = timeouts -> next;
	      (*(t -> func)) (t -> what);
	      t -> next = free_timeouts;
	      free_timeouts = t;
	      goto another;
	    }
	  tv.tv_sec = timeouts -> when - cur_time;
	  tv.tv_usec = 0;
	  tvp = &tv;
	}
      else
	tvp = (struct timeval *)0;

      /* Set up the read mask. */
      FD_ZERO (&r);

      for (l = protocols; l; l = l -> next)
	{
	  FD_SET (l -> fd, &r);
	  if (l -> fd > max)
	    max = l -> fd;
	}

      /* Wait for a packet or a timeout... XXX */
      count = select (max + 1, &r, &w, &x, tvp);

      /* Get the current time... */
      GET_TIME (&cur_time);

      /* Not likely to be transitory... */
      if (count < 0)
	error ("select: %m");

      for (l = protocols; l; l = l -> next)
	{
	  if (!FD_ISSET (l -> fd, &r))
	    continue;
	  if (l -> handler)
	    (*(l -> handler)) (l);
	  if (interfaces_invalidated)
	    break;
	}
      interfaces_invalidated = 0;
    }
  while (1);
}

static void 
got_one(struct protocol *l)
{
  struct sockaddr_in from;
  struct hardware hfrom;
  int result;
  static u_char packbuf [4095]; /* Packet input buffer.
				   Must be as large as largest
				   possible MTU. */
  struct interface_info *ip = l -> local;

  if ((result = receive_packet (ip, packbuf, sizeof packbuf, 
				&from, &hfrom)) < 0)
    {
      warn ("receive_packet failed on %s: %m", ip -> name);
      return;
    }
  if (result == 0)
    return;

  if (fumied_packet_handler)
    {
      (*fumied_packet_handler) (ip, packbuf, result, from.sin_port, &hfrom);
    }
}

/* Add a protocol to the list of protocols... */
void
add_protocol(char *name, int fd, void (*handler)(), void *local)
{
  struct protocol *p;

  p = (struct protocol *)malloc (sizeof *p);
  if (!p)
    error ("can't allocate protocol struct for %s", name);

  p -> fd = fd;
  p -> handler = handler;
  p -> local = local;

  p -> next = protocols;
  protocols = p;
}

void
do_packet(struct interface_info *interface, u_char *packbuf, int len,
	  unsigned short from_port, struct hardware *hfrom)
{
  struct packet tp;
  struct fumiedMessage tdp;
  u_long cookie;

  memcpy (&tdp, packbuf, len);
  memset (&tp, 0, sizeof tp);
  tp.raw = &tdp;
  tp.packet_length = len;
  tp.client_port = from_port;
  tp.interface = interface;
  tp.haddr = hfrom;

  cookie = htonl(MAGIC_COOKIE);
  if (memcmp (tp.raw->options, &cookie, 4))
    {
      return;
    }

  if ((tp.raw->options[4] == FUMIEDDISCOVER) ||
      (tp.raw->options[4] == FUMIEDREQUEST))
    {
      tp.packet_type = tp.raw->options[4];
      tp.raw->siaddr = interface->primary_address;
      if (tp.packet_type)
	fumied (&tp);
    }
}
