/* fumied.h */

#include <syslog.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <net/if.h>
#include <sys/time.h>
#include <stdarg.h>

#define	N_REXMIT_REQUEST	2
#define	FUMIED_SERVER_PORT 4126	/* from client to server */
#define	FUMIED_CLIENT_PORT 4127	/* from server to client */

#define FUMIED_UDP_OVERHEAD	(14 + /* Ethernet header */		\
				 20 + /* IP header */			\
				 8)   /* UDP header */
#define FUMIED_FIXED_NON_UDP	236
#define FUMIED_FIXED_LEN	(FUMIED_FIXED_NON_UDP + FUMIED_UDP_OVERHEAD)
#define FUMIED_MTU_MAX		1500
#define FUMIED_OPTION_LEN	(FUMIED_MTU_MAX - FUMIED_FIXED_LEN)
#define FUMIED_MIN_LEN            548

struct fumiedMessage {
  u_char  htype;		/* Hardware addr type (see net/if_types.h) */
  u_char  hlen;			/* Hardware addr length */
  u_int flags;			/* Flag bits */
  struct in_addr ciaddr;	/* Client IP address (if already in use) */
  struct in_addr siaddr;	/* IP address of next server to talk to */
  u_char options [FUMIED_OPTION_LEN];
};

struct ifinfo {
  char ifname[IFNAMSIZ];
  u_long addr;			/* network byte order */
  u_long mask;			/* network byte order */
  u_long bcast;			/* network byte order */
  short	flags;
};

/* FUMIED client states
 */
enum {
  MAX_STATES	= 4,		/* number of states */
  INIT		= 0,
  REQUESTING	= 1,
  GETTING	= 2,
  COMPLETED	= 3,
  EXCEPTION	= -1
};

/* Timeout control
 */
enum {
  INIT_TIMEOUT = 0,
  NEXT_TIMEOUT = 1,
};

#define MAGIC_COOKIE	0x36352836	/* magic cookie in the option field */

/* Possible values for hardware type (htype) field... */
#define HTYPE_ETHER	1               /* Ethernet 10Mbps              */
#define HTYPE_IEEE802	6               /* IEEE 802.2 Token Ring...	*/

/* FUMIED message types. */
#define FUMIEDDISCOVER	1
#define FUMIEDREQUEST	2
#define FUMIEDACK	3
#define FUMIEDNAK	4

#define ARPHRD_ETHER 1
#define TIME time_t
#define GET_TIME(x)	time ((x))

struct packet {
  struct fumiedMessage *raw;
  int packet_length;
  int packet_type;
  int options_valid;
  int client_port;
  struct interface_info *interface;	/* Interface on which packet
					   was received. */
  struct hardware *haddr;		/* Physical link address
					   of local sender (maybe gateway). */
} packet;

struct hardware {
  u_char htype;
  u_char hlen;
  u_char haddr [16];
};

#define DISCOVER_RUNNING	0
#define DISCOVER_SERVER		1
#define DISCOVER_UNCONFIGURED	2
#define DISCOVER_RELAY		3
#define DISCOVER_REQUESTED	4

struct interface_info {
  struct interface_info *next;	/* Next interface in list... */
  struct shared_network *shared_network;
				/* Networks connected to this interface. */
  struct hardware hw_address;	/* Its physical address. */
  struct in_addr primary_address;	/* Primary interface address. */
  char name [16];		/* Its name... */
  int rfdesc;			/* Its read file descriptor. */
  int wfdesc;			/* Its write file descriptor, if
				   different. */
  u_char *rbuf;			/* Read buffer, if required. */
  size_t rbuf_max;		/* Size of read buffer. */
  size_t rbuf_offset;		/* Current offset into buffer. */
  size_t rbuf_len;		/* Length of data in buffer. */
  
  struct ifreq *ifp;		/* Pointer to ifreq struct. */
  u_long flags;			/* Control flags... */
#define INTERFACE_REQUESTED 1
#define INTERFACE_AUTOMATIC 2
};

struct timeout {
  struct timeout *next;
  TIME when;
  void (*func)();
  void *what;
};

struct protocol {
  struct protocol *next;
  int fd;
  void (*handler)();
  void *local;
};

#define _SUID_PATH_FUMIED_PID	"/var/run/fumied.pid"
#define _SUID_ETC_HOSTS		"/etc/hosts"
#define _SUID_TEMP_HOSTS	"/etc/hosts.XXXXXX"
#define _PATH_FUMIED_PID	"/.fumied.pid"
#define _ETC_HOSTS		"/hosts"
#define _TEMP_HOSTS		"/hosts.XXXXXX"
#define FUMIED_LOG_FACILITY	LOG_DAEMON

/* mian.c */
extern char fumied_pid_path[128];
extern char etc_hosts[128];
extern char temp_hosts[128];
int main(int argc, char **argv);
void cleanup();

/* errwarn.c */
void error(char *fmt, ...);
int warn(char *fmt, ...);
int note(char *fmt, ...);
int debug(char *fmt, ...);

/* fumied.c */
extern TIME cur_time;
extern u_int local_port;
extern u_int remote_port;
extern int log_priority;
extern int log_perror;
extern struct ifinfo Ifbuf;
void fumied(struct packet *packet);
void fumieddiscover(struct packet *packet);
void fumiedrequest(struct packet *packet);
void revise_hosts (char *local_hostname, char *local_ipaddr,
		   char *peer_hostname, char *peer_ipaddr);

/* alloc.c */
void *dmalloc(int size, char *name);
void dfree(void *ptr, char *name);

/* socket-if.c */
void getIfInfo(char *ifname);
void setSockAddrIn(u_short port, u_long inaddr, struct sockaddr_in *saddr);
void openSendSocket(struct sockaddr_in *addr, int *s);
void openRecvSocket(struct sockaddr_in *addr, int *s);
int rcvAndCheckFumiedMsg(int s, struct fumiedMessage *msg, u_long waitMsgType,
			long timeout);
int waitChkReXmitMsg(int sRecv, struct fumiedMessage *pMsgRecv,
		     int sSend, struct fumiedMessage *pMsgSend,
		     struct sockaddr_in *addr, u_long waitMsgType, int nretry);
void setWaitMsgType(int type, u_int *ptype);
int if_register_socket(struct interface_info *info);
void if_register_send(struct interface_info *info);
void if_register_receive(struct interface_info *info);
ssize_t send_packet(struct interface_info *interface, struct packet *packet,
		    struct fumiedMessage *raw, size_t len, struct in_addr from,
		    struct sockaddr_in *to, struct hardware *hto);
ssize_t receive_packet(struct interface_info *interface, u_char *buf,
		       size_t len, struct sockaddr_in *from,
		       struct hardware *hfrom);

/* client.h */
void fumiedMsgInit(u_char *ifname);
void fumiedClient();
int init();
int request();
int getinfo();
void mkFumiedDiscoverMsg(struct fumiedMessage *msg);
void mkFumiedRequestMsg(struct fumiedMessage *msg);
long getNextTimeout(int flag);

/* dispatch.c */
extern struct interface_info *interfaces, *dummy_interfaces;
extern struct protocol *protocols;
extern struct timeout *timeouts;
extern void (*fumied_packet_handler)(struct interface_info *interface,
				    u_char *packbuf, int len,
				    unsigned short from_port,
				    struct hardware *hfrom);
void discover_interfaces(int state);
void dispatch();
void add_protocol(char *name, int fd, void (*handler)(), void *local);
void do_packet(struct interface_info *interface, u_char *packbuf, int len,
	       unsigned short from_port, struct hardware *hfrom);
